module Quox.Syntax.Term

import public Quox.Syntax.Var
import public Quox.Syntax.Shift
import public Quox.Syntax.Subst
import public Quox.Syntax.Universe
import public Quox.Syntax.Qty
import public Quox.Syntax.Dim
import public Quox.Name

import Quox.Pretty

import public Data.DPair
import Data.List
import Data.Maybe
import Data.Nat
import public Data.So
import Data.String
import Data.Vect

%default total


infixl 8 :#
infixl 9 :@
mutual
  public export
  TSubst : Nat -> Nat -> Nat -> Type
  TSubst d = Subst (\n => Elim d n)

  ||| first argument `d` is dimension scope size, second `n` is term scope size
  public export
  data Term : (d, n : Nat) -> Type where
    ||| type of types
    TYPE : (l : Universe) -> Term d n

    ||| function type
    Pi  : (qtm, qty : Qty) -> (x : Name) ->
          (arg : Term d n) -> (res : Term d (S n)) -> Term d n
    ||| function term
    Lam : (x : Name) -> (body : Term d (S n)) -> Term d n

    ||| elimination
    E : (e : Elim d n) -> Term d n

    ||| term closure/suspended substitution
    CloT  : (tm : Term d from)  -> (th : Lazy (TSubst d from to)) -> Term d to
    ||| dimension closure/suspended substitution
    DCloT : (tm : Term dfrom n) -> (th : Lazy (DSubst dfrom dto)) -> Term dto n

  ||| first argument `d` is dimension scope size, second `n` is term scope size
  public export
  data Elim : (d, n : Nat) -> Type where
    ||| free variable
    F : (x : Name) -> Elim d n
    ||| bound variable
    B : (i : Var n) -> Elim d n

    ||| term application
    (:@) : (fun : Elim d n) -> (arg : Term d n) -> Elim d n

    ||| type-annotated term
    (:#) : (tm, ty : Term d n) -> Elim d n

    ||| term closure/suspended substitution
    CloE  : (el : Elim d from)  -> (th : Lazy (TSubst d from to)) -> Elim d to
    ||| dimension closure/suspended substitution
    DCloE : (el : Elim dfrom n) -> (th : Lazy (DSubst dfrom dto)) -> Elim dto n

%name Term s, t, r
%name Elim e, f


||| same as `F` but as a term
public export %inline
FT : Name -> Term d n
FT = E . F

||| abbreviation for a bound variable like `BV 4` instead of
||| `B (VS (VS (VS (VS VZ))))`
public export %inline
BV : (i : Nat) -> (0 _ : LT i n) => Elim d n
BV i = B $ V i

||| same as `BV` but as a term
public export %inline
BVT : (i : Nat) -> (0 _ : LT i n) => Term d n
BVT i = E $ BV i


public export %inline
isLam : Term d n -> Bool
isLam (Lam {}) = True
isLam _        = False

public export
NotLam : Term d n -> Type
NotLam = So . not . isLam


public export %inline
isApp : Elim d n -> Bool
isApp ((:@) {}) = True
isApp _         = False

public export
NotApp : Elim d n -> Type
NotApp = So . not . isApp


infixl 9 :@@
||| apply multiple arguments at once
public export %inline
(:@@) : Elim d n -> List (Term d n) -> Elim d n
f :@@ ss = foldl (:@) f ss

public export
record GetArgs (d, n : Nat) where
  constructor GotArgs
  fun      : Elim d n
  args     : List (Term d n)
  0 notApp : NotApp fun

private
getArgs' : Elim d n -> List (Term d n) -> GetArgs d n
getArgs' fun args with (choose $ isApp fun)
  getArgs' (f :@ a) args | Left yes = getArgs' f (a :: args)
  _ | Right no = GotArgs {fun, args, notApp = no}

||| splits an application into its head and arguments. if it's not an
||| application then the list is just empty
export %inline
getArgs : Elim d n -> GetArgs d n
getArgs e = getArgs' e []


infixr 1 :\\
public export
(:\\) : Vect m Name -> Term d (m + n) -> Term d n
[]      :\\ t = t
x :: xs :\\ t = let t' = replace {p = Term _} (plusSuccRightSucc {}) t in
                Lam x $ xs :\\ t'

public export
record GetLams (d, n : Nat) where
  constructor GotLams
  names    : Vect lams Name
  body     : Term d rest
  0 eq     : lams + n = rest
  0 notLam : NotLam body

public export
getLams : Term d n -> GetLams d n
getLams s with (choose $ isLam s)
  getLams (Lam x s) | Left yes =
    let inner = getLams s in
    GotLams {names  = x :: inner.names,
             body   = inner.body,
             eq     = plusSuccRightSucc {} `trans` inner.eq,
             notLam = inner.notLam}
  _ | Right no = GotLams {names = [], body = s, eq = Refl, notLam = no}


parameters {auto _ : Pretty.HasEnv m}
  private %inline arrowD : m (Doc HL)
  arrowD = hlF Syntax $ ifUnicode "→" "->"

  private %inline lamD : m (Doc HL)
  lamD = hlF Syntax $ ifUnicode "λ" "fun"

  private %inline annD : m (Doc HL)
  annD = hlF Syntax $ ifUnicode "⦂" "::"

private %inline typeD : Doc HL
typeD = hl Syntax "Type"

private %inline colonD : Doc HL
colonD = hl Syntax ":"

mutual
  export covering
  PrettyHL (Term d n) where
    prettyM (TYPE l) =
      parensIfM App $ typeD <//> !(withPrec Arg $ prettyM l)
    prettyM (Pi qtm qty x s t) =
      parensIfM Outer $ hang 2 $
        !(prettyBinder [qtm, qty] x s) <++> !arrowD
          <//> !(under T x $ prettyM t)
    prettyM (Lam x t) =
      parensIfM Outer $
        sep [!lamD, hl TVar !(prettyM x), !arrowD]
          <//> !(under T x $ prettyM t)
    prettyM (E e) =
      prettyM e
    prettyM (CloT s th) =
      parensIfM SApp . hang 2 =<<
        [|withPrec SApp (prettyM s) </> prettyTSubst th|]
    prettyM (DCloT s th) =
      parensIfM SApp . hang 2 =<<
        [|withPrec SApp (prettyM s) </> prettyDSubst th|]

  export covering
  PrettyHL (Elim d n) where
    prettyM (F x) =
      hl' Free <$> prettyM x
    prettyM (B i) =
      prettyVar TVar TVarErr (!ask).tnames i
    prettyM (e :@ s) =
      let GotArgs f args _ = getArgs' e [s] in
      parensIfM App =<< withPrec Arg
        [|prettyM f <//> (align . sep <$> traverse prettyM args)|]
    prettyM (s :# a) =
      parensIfM Ann $ hang 2 $
        !(withPrec AnnL $ prettyM s) <++> !annD
          <//> !(withPrec Ann $ prettyM a)
    prettyM (CloE e th)  =
      parensIfM SApp . hang 2 =<<
        [|withPrec SApp (prettyM e) </> prettyTSubst th|]
    prettyM (DCloE e th) =
      parensIfM SApp . hang 2 =<<
        [|withPrec SApp (prettyM e) </> prettyDSubst th|]

  export covering
  prettyTSubst : Pretty.HasEnv m => TSubst d from to -> m (Doc HL)
  prettyTSubst s = prettySubstM prettyM (!ask).tnames TVar "[" "]" s

  export covering
  prettyBinder : Pretty.HasEnv m => List Qty -> Name -> Term d n -> m (Doc HL)
  prettyBinder pis x a =
    pure $ parens $ hang 2 $
      hsep [hl TVar !(prettyM x),
            sep [!(prettyQtyBinds pis),
                 hsep [colonD, !(withPrec Outer $ prettyM a)]]]


export FromVar (Elim d) where fromVar = B
export FromVar (Term d) where fromVar = E . fromVar

||| does the minimal reasonable work:
||| - deletes the closure around a free name since it doesn't do anything
||| - deletes an identity substitution
||| - composes (lazily) with an existing top-level closure
||| - immediately looks up a bound variable
||| - otherwise, wraps in a new closure
export
CanSubst (Elim d) (Elim d) where
  F x       // _  = F x
  B i       // th = th !! i
  CloE e ph // th = assert_total CloE e $ ph . th
  e         // th = case force th of
                         Shift SZ => e
                         th       => CloE e th

||| does the minimal reasonable work:
||| - deletes the closure around an atomic constant like `TYPE`
||| - deletes an identity substitution
||| - composes (lazily) with an existing top-level closure
||| - goes inside `E` in case it is a simple variable or something
||| - otherwise, wraps in a new closure
export
CanSubst (Elim d) (Term d) where
  TYPE l    // _  = TYPE l
  E e       // th = E $ e // th
  CloT s ph // th = CloT s $ ph . th
  s         // th = case force th of
                         Shift SZ => s
                         th       => CloT s th

export CanSubst Var (Term d) where s // th = s // map (B {d}) th
export CanSubst Var (Elim d) where e // th = e // map (B {d}) th

export CanShift (Term d) where i // by = i // Shift by {env = Elim d}
export CanShift (Elim d) where i // by = i // Shift by {env = Elim d}


infixl 8 ///
mutual
  namespace Term
    ||| applies a dimension substitution with the same behaviour as `(//)`
    ||| above
    export
    (///) : Term dfrom n -> DSubst dfrom dto -> Term dto n
    TYPE l     /// _        = TYPE l
    E e        /// th       = E $ e /// th
    DCloT s ph /// th       = DCloT s $ ph . th
    s          /// Shift SZ = s
    s          /// th       = DCloT s th

    ||| applies a term and dimension substitution
    public export %inline
    subs : Term dfrom from -> DSubst dfrom dto -> TSubst dto from to ->
           Term dto to
    subs s th ph = s /// th // ph

  namespace Elim
    ||| applies a dimension substitution with the same behaviour as `(//)`
    ||| above
    export
    (///) : Elim dfrom n -> DSubst dfrom dto -> Elim dto n
    F x        /// _        = F x
    B i        /// _        = B i
    DCloE e ph /// th       = DCloE e $ ph . th
    e          /// Shift SZ = e
    e          /// th       = DCloE e th

    ||| applies a term and dimension substitution
    public export %inline
    subs : Elim dfrom from -> DSubst dfrom dto -> TSubst dto from to ->
           Elim dto to
    subs e th ph = e /// th // ph


private %inline
comp' : DSubst dfrom dto -> TSubst dfrom from mid -> TSubst dto mid to ->
        TSubst dto from to
comp' th ps ph = map (/// th) ps . ph


mutual
  ||| true if a term has a closure or dimension closure at the top level,
  ||| or is `E` applied to such an elimination
  public export %inline
  topCloT : Term d n -> Bool
  topCloT (CloT  _ _) = True
  topCloT (DCloT _ _) = True
  topCloT (E e)       = topCloE e
  topCloT _           = False

  ||| true if an elimination has a closure or dimension closure at the top level
  public export %inline
  topCloE : Elim d n -> Bool
  topCloE (CloE  _ _) = True
  topCloE (DCloE _ _) = True
  topCloE _           = False


public export IsNotCloT : Term d n -> Type
IsNotCloT = So . not . topCloT

||| a term which is not a top level closure
public export NotCloTerm : Nat -> Nat -> Type
NotCloTerm d n = Subset (Term d n) IsNotCloT

public export IsNotCloE : Elim d n -> Type
IsNotCloE = So . not . topCloE

||| an elimination which is not a top level closure
public export NotCloElim : Nat -> Nat -> Type
NotCloElim d n = Subset (Elim d n) IsNotCloE

public export %inline
ncloT : (t : Term d n) -> (0 _ : IsNotCloT t) => NotCloTerm d n
ncloT t @{p} = Element t p

public export %inline
ncloE : (t : Elim d n) -> (0 _ : IsNotCloE t) => NotCloElim d n
ncloE e @{p} = Element e p



mutual
  ||| if the input term has any top-level closures, push them under one layer of
  ||| syntax
  export %inline
  pushSubstsT : Term d n -> NotCloTerm d n
  pushSubstsT s = pushSubstsTWith id id s

  ||| if the input elimination has any top-level closures, push them under one
  ||| layer of syntax
  export %inline
  pushSubstsE : Elim d n -> NotCloElim d n
  pushSubstsE e = pushSubstsEWith id id e

  export
  pushSubstsTWith : DSubst dfrom dto -> TSubst dto from to ->
                    Term dfrom from -> NotCloTerm dto to
  pushSubstsTWith th ph (TYPE l) =
    ncloT $ TYPE l
  pushSubstsTWith th ph (Pi qtm qty x a b) =
    ncloT $ Pi qtm qty x (subs a th ph) (subs b th (push ph))
  pushSubstsTWith th ph (Lam x t) =
    ncloT $ Lam x $ subs t th $ push ph
  pushSubstsTWith th ph (E e) =
    let Element e _ = pushSubstsEWith th ph e in ncloT $ E e
  pushSubstsTWith th ph (CloT  s ps) =
    pushSubstsTWith th (comp' th ps ph) s
  pushSubstsTWith th ph (DCloT s ps) =
    pushSubstsTWith (ps . th) ph s

  export
  pushSubstsEWith : DSubst dfrom dto -> TSubst dto from to ->
                    Elim dfrom from -> NotCloElim dto to
  pushSubstsEWith th ph (F x) =
    ncloE $ F x
  pushSubstsEWith th ph (B i) =
    assert_total pushSubstsE $ ph !! i
  pushSubstsEWith th ph (f :@ s) =
    ncloE $ subs f th ph :@ subs s th ph
  pushSubstsEWith th ph (s :# a) =
    ncloE $ subs s th ph :# subs a th ph
  pushSubstsEWith th ph (CloE  e ps) =
    pushSubstsEWith th (comp' th ps ph) e
  pushSubstsEWith th ph (DCloE e ps) =
    pushSubstsEWith (ps . th) ph e


parameters (th : DSubst dfrom dto) (ph : TSubst dto from to)
  public export %inline
  pushSubstsTWith' : Term dfrom from -> Term dto to
  pushSubstsTWith' s = (pushSubstsTWith th ph s).fst

  public export %inline
  pushSubstsEWith' : Elim dfrom from -> Elim dto to
  pushSubstsEWith' e = (pushSubstsEWith th ph e).fst


public export %inline
pushSubstsT' : Term d n -> Term d n
pushSubstsT' s = (pushSubstsT s).fst

public export %inline
pushSubstsE' : Elim d n -> Elim d n
pushSubstsE' e = (pushSubstsE e).fst


public export
data IsRedexT : Term d n -> Type where
  IsUpsilon : IsRedexT $ E (_ :# _)
  IsCloT    : IsRedexT $ CloT {}
  IsDCloT   : IsRedexT $ DCloT {}

public export %inline
NotRedexT : Term d n -> Type
NotRedexT = Not . IsRedexT

public export
data IsRedexE : Elim d n -> Type where
  IsBetaLam : IsRedexE $ (Lam {} :# Pi {}) :@ _
  IsCloE    : IsRedexE $ CloE {}
  IsDCloE   : IsRedexE $ DCloE {}

public export %inline
NotRedexE : Elim d n -> Type
NotRedexE = Not . IsRedexE


export %inline
isRedexT : (t : Term d n) -> Dec (IsRedexT t)
isRedexT (E (_ :# _))   = Yes IsUpsilon
isRedexT (CloT {})      = Yes IsCloT
isRedexT (DCloT {})     = Yes IsDCloT
isRedexT (TYPE _)       = No $ \x => case x of {}
isRedexT (Pi {})        = No $ \x => case x of {}
isRedexT (Lam {})       = No $ \x => case x of {}
isRedexT (E (F _))      = No $ \x => case x of {}
isRedexT (E (B _))      = No $ \x => case x of {}
isRedexT (E (_ :@ _))   = No $ \x => case x of {}
isRedexT (E (CloE {}))  = No $ \x => case x of {}
isRedexT (E (DCloE {})) = No $ \x => case x of {}

export %inline
isRedexE : (e : Elim d n) -> Dec (IsRedexE e)
isRedexE ((Lam {} :# Pi {}) :@ _)      = Yes IsBetaLam
isRedexE (CloE {})                     = Yes IsCloE
isRedexE (DCloE {})                    = Yes IsDCloE
isRedexE (F x)                         = No $ \x => case x of {}
isRedexE (B i)                         = No $ \x => case x of {}
isRedexE (F _                    :@ _) = No $ \x => case x of {}
isRedexE (B _                    :@ _) = No $ \x => case x of {}
isRedexE (_ :@ _                 :@ _) = No $ \x => case x of {}
isRedexE ((TYPE _   :# _)        :@ _) = No $ \x => case x of {}
isRedexE ((Pi {}    :# _)        :@ _) = No $ \x => case x of {}
isRedexE ((Lam {}   :# TYPE _)   :@ _) = No $ \x => case x of {}
isRedexE ((Lam {}   :# Lam {})   :@ _) = No $ \x => case x of {}
isRedexE ((Lam {}   :# E _)      :@ _) = No $ \x => case x of {}
isRedexE ((Lam {}   :# CloT {})  :@ _) = No $ \x => case x of {}
isRedexE ((Lam {}   :# DCloT {}) :@ _) = No $ \x => case x of {}
isRedexE ((E _      :# _)        :@ _) = No $ \x => case x of {}
isRedexE ((CloT {}  :# _)        :@ _) = No $ \x => case x of {}
isRedexE ((DCloT {} :# _)        :@ _) = No $ \x => case x of {}
isRedexE (CloE {}                :@ _) = No $ \x => case x of {}
isRedexE (DCloE {}               :@ _) = No $ \x => case x of {}
isRedexE (_ :# _)                      = No $ \x => case x of {}


public export %inline
RedexTerm : Nat -> Nat -> Type
RedexTerm d n = Subset (Term d n) IsRedexT

public export %inline
NonRedexTerm : Nat -> Nat -> Type
NonRedexTerm d n = Subset (Term d n) NotRedexT


public export %inline
RedexElim : Nat -> Nat -> Type
RedexElim d n = Subset (Elim d n) IsRedexE

public export %inline
NonRedexElim : Nat -> Nat -> Type
NonRedexElim d n = Subset (Elim d n) NotRedexE


public export %inline
stepT' : (s : Term d n) -> IsRedexT s -> Term d n
stepT' (E (s :# _)) IsUpsilon = s
stepT' (CloT  s th) IsCloT    = pushSubstsTWith' id th s
stepT' (DCloT s th) IsDCloT   = pushSubstsTWith' th id s

public export %inline
stepT : (s : Term d n) -> Either (NotRedexT s) (Term d n)
stepT s = case isRedexT s of Yes y => Right $ stepT' s y; No n => Left n

public export %inline
stepE' : (e : Elim d n) -> IsRedexE e -> Elim d n
stepE' ((Lam {body, _} :# Pi {arg, res, _}) :@ s) IsBetaLam =
  (body :# res) // one (s :# arg)
stepE' (CloE  e th) IsCloE  = pushSubstsEWith' id th e
stepE' (DCloE e th) IsDCloE = pushSubstsEWith' th id e

public export %inline
stepE : (e : Elim d n) -> Either (NotRedexE e) (Elim d n)
stepE e = case isRedexE e of Yes y => Right $ stepE' e y; No n => Left n

public export
whnfT : Term d n -> NonRedexTerm d n
whnfT s = case stepT s of
               Right s'   => whnfT $ assert_smaller s s'
               Left  done => Element s done

public export
whnfE : Elim d n -> NonRedexElim d n
whnfE e = case stepE e of
               Right e'   => whnfE $ assert_smaller e e'
               Left  done => Element e done


public export
Exists2 : (ty1 -> ty2 -> Type) -> Type
Exists2 t = Exists $ Exists . t

public export %inline
some : {0 f : a -> Type} -> f x -> Exists f
some p = Evidence _ p

public export %inline
some2 : {0 f : a -> b -> Type} -> f x y -> Exists2 f
some2 p = some $ some p


||| A term of some unknown scope sizes.
public export
SomeTerm : Type
SomeTerm = Exists2 Term

||| An elimination of some unknown scope sizes.
public export
SomeElim : Type
SomeElim = Exists2 Elim

||| A dimension of some unknown scope size.
public export
SomeDim : Type
SomeDim = Exists Dim
